/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.ftp.connection;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.activation.FileTypeMap;
import javax.mail.util.ByteArrayDataSource;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.net.ftp.FTP;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPFile;
import org.ow2.petals.bc.ftp.FTPClientPool;
import org.ow2.petals.component.framework.util.FileNamePatternUtil;
import org.ow2.petals.component.framework.util.FileNamePatternUtil.RegexFileNameFilter;

/**
 * Some IO utilities for the file transfer BC.
 * 
 * @version $Rev: 250 $Date: {date}
 * @since Petals 1.2
 * @author alouis
 * 
 */
public final class FTPConnection {

    private Logger logger;

    private FTPConnectionInfo info;

    public FTPConnection(FTPConnectionInfo info, Logger logger) throws IOException {
        super();
        this.info = info;
        this.logger = Logger.getLogger(logger.getName() + ".FTPConnection");
    }

    /**
     * Put the file on the FTP server.
     * 
     * @param fileName
     * @param content
     * @return
     * @throws IOException
     * @throws InterruptedException
     */
    public void putString(final String fileName, String content) throws IOException,
            InterruptedException {
        FTPClient client = null;
        boolean transmit = false;
        boolean fileExist = false;
        int numberOfTries = 0;
        int maxTries = info.getAttempt();

        try {
            while (!transmit) {
                numberOfTries++;

                if (numberOfTries > maxTries) {
                    throw new IOException("Error occurred while storing '" + fileName
                            + "' file on the FTP server");
                }

                try {
                    logger.finest("start sending '" + fileName + "' file on the FTP server");

                    if (client == null || !client.isConnected()) {
                        client = connectToFTPDirectory();
                    }
                    fileExist = this.fileExistOnRemoteDirectory(fileName, client);
                    if (!fileExist || this.info.isOverwrite()) {
                        OutputStream ops = client.storeFileStream(fileName);
                        ops.write(content.getBytes());
                        ops.flush();
                        ops.close();
                        transmit = client.completePendingCommand();
                    } else {
                        // hack to
                        numberOfTries = maxTries;
                        throw new IOException("File [" + fileName + "] already exists");
                    }
                } catch (IOException e) {
                    if (numberOfTries >= maxTries) {
                        throw e;
                    } else {
                        this.logger.warning("An attempt to put the file '" + fileName
                                + "' failed : " + e.getMessage());
                    }
                    Thread.sleep(info.getDelay());
                }
            }
        } finally {
            if (client != null && client.isConnected()) {
                logger.finest("disconnect FTP connection to " + info.getServer());
                client.logout();
                client.disconnect();
            }
        }
        logger.finest("'" + fileName + "' sent on the FTP server");
    }

    public void del(String remoteFile) throws IOException, InterruptedException {
        FTPClient client = null;
        int numberOfTries = 0;
        int maxTries = info.getAttempt();
        boolean deleted = false;
        try {
            while (!deleted) {
                numberOfTries++;
                if (numberOfTries > maxTries) {
                    throw new IOException("Error occurred while deleting '" + remoteFile
                            + "' file from the FTP server");
                }

                try {
                    if (this.logger.isLoggable(Level.FINEST)) {
                        logger.finest("start deleting '" + remoteFile
                                + "' file from the FTP server");
                    }

                    if (client == null || !client.isConnected()) {
                        client = connectToFTPDirectory();
                    }
                    deleted = client.deleteFile(remoteFile);
                } catch (IOException e) {
                    if (numberOfTries == maxTries) {
                        throw e;
                    } else {
                        this.logger.warning("An attempt to delete the file '" + remoteFile
                                + "' failed : " + e.getMessage());
                    }
                    Thread.sleep(info.getDelay());
                }
            }
        } finally {
            if (client != null && client.isConnected()) {
                if (this.logger.isLoggable(Level.FINEST)) {
                    this.logger.finest("disconnect FTP connection to " + info.getServer());
                }
                client.logout();
                client.disconnect();
            }
        }
    }

    public void mDel(List<String> remoteFiles) throws IOException, InterruptedException {
        FTPClient client = null;
        int numberOfTries = 0;
        int maxTries = info.getAttempt();
        boolean deletedAllFile = false;
        try {
            while (!deletedAllFile) {
                try {
                    if (this.logger.isLoggable(Level.FINEST)) {
                        logger.finest("start deleting '" + remoteFiles.toString()
                                + "' file from the FTP server");
                    }

                    if (client == null || !client.isConnected()) {
                        client = connectToFTPDirectory();
                    }

                    for (String pathname : remoteFiles) {
                        client.deleteFile(pathname);
                    }
                    deletedAllFile = true;
                } catch (IOException e) {
                    if (numberOfTries == maxTries) {
                        throw e;
                    } else {
                        this.logger.warning("An attempt to delete the file '" + remoteFiles
                                + "' failed : " + e.getMessage());
                    }
                    Thread.sleep(info.getDelay());
                }
            }
        } finally {
            if (client != null && client.isConnected()) {
                if (this.logger.isLoggable(Level.FINEST)) {
                    this.logger.finest("disconnect FTP connection to " + info.getServer());
                }
                client.logout();
                client.disconnect();
            }
        }
    }

    /**
     * Put files on the FTP server.
     * 
     * @param attachments
     * @return
     * @throws IOException
     * @throws InterruptedException
     */
    public void mPut(Map<String, DataHandler> attachments) throws IOException, InterruptedException {
        FTPClient client = null;

        try {
            int numberOfTries = 0;
            int maxTries = info.getAttempt();
            boolean transmit;
            boolean isFileExist;
            for (String fileName : attachments.keySet()) {
                transmit = false;
                isFileExist = false;
                while (!transmit) {

                    if (numberOfTries == maxTries) {
                        throw new IOException("Error occurred while storing '" + fileName
                                + "' file on the FTP server");
                    }

                    try {
                        logger.finest("start sending '" + fileName + "' file on the FTP server");

                        if (client == null || !client.isConnected()) {
                            client = connectToFTPDirectory();
                        }
                        isFileExist = this.fileExistOnRemoteDirectory(fileName, client);
                        if (!isFileExist || this.info.isOverwrite()) {
                            InputStream stream = attachments.get(fileName).getInputStream();
                            transmit = client.storeFile(fileName, stream);
                            stream.close();
                        } else {
                            numberOfTries = maxTries;
                            throw new IOException("File [" + fileName + "] already exists");
                        }
                        if (transmit) {
                            logger.finest("'" + fileName + "' sent on the FTP server");
                        } else {
                            numberOfTries++;
                        }
                    } catch (IOException e) {
                        numberOfTries++;
                        if (numberOfTries >= maxTries) {
                            throw e;
                        } else {
                            this.logger.warning("An attempt to put the file '" + fileName
                                    + "' failed : " + e.getMessage());
                        }
                        Thread.sleep(info.getDelay());
                    }
                }
            }
        } finally {
            if (client != null && client.isConnected()) {
                logger.finest("disconnect FTP connection to " + info.getServer());
                client.logout();
                client.disconnect();
            }
        }
    }

    /**
     * Gets a file from a FTP server.
     * 
     * @param server
     *            the server
     * @param port
     *            the port
     * @param login
     *            the login
     * @param password
     *            the password
     * @param remoteFile
     *            the remote file
     * 
     * @return String
     * @throws InterruptedException
     * @throws FTPBCException
     *             impossible to get: The file does not exist
     */
    public Source get(final String remoteFile) throws IOException, InterruptedException {
        Source result = null;
        FTPClient client = null;
        int numberOfTries = 0;
        int maxTries = info.getAttempt();

        try {
            while (result == null) {
                numberOfTries++;

                if (numberOfTries > maxTries) {
                    throw new IOException("Error occurred while getting '" + remoteFile
                            + "' file from the FTP server");
                }

                try {
                    logger.finest("start getting '" + remoteFile + "' file from the FTP server");

                    if (client == null || !client.isConnected()) {
                        client = connectToFTPDirectory();
                    }

                    FTPFile[] files = client.listFiles(remoteFile);
                    if (files != null && files.length > 0) {
                        result = getAsSource(files[0], client);
                    }
                } catch (IOException e) {
                    if (numberOfTries == maxTries) {
                        throw e;
                    } else {
                        this.logger.warning("An attempt to get the file '" + remoteFile
                                + "' failed : " + e.getMessage());
                    }
                    Thread.sleep(info.getDelay());
                }
            }
        } finally {
            if (client != null && client.isConnected()) {
                logger.finest("disconnect FTP connection to " + info.getServer());
                client.logout();
                client.disconnect();
            }
        }

        return result;
    }

    /**
     * This method will only get files, it does not include subpath
     * 
     * @param fileFilterList
     * @return
     * @throws IOException
     * @throws InterruptedException
     */
    public Map<String, DataHandler> mGet(final List<String> fileFilterList) throws IOException,
            InterruptedException {
        FTPClient client = null;
        boolean getAllFiles = false;
        int numberOfTries = 0;
        int maxTries = info.getAttempt();
        Map<String, DataHandler> datahandlers = new HashMap<String, DataHandler>();

        // Construct the fileFilters to match files read from the FTP directory
        List<FileNamePatternUtil.RegexFileNameFilter> filterList = new LinkedList<FileNamePatternUtil.RegexFileNameFilter>();
        for (String string : fileFilterList) {
            filterList.add(FileNamePatternUtil.getInstance()
                    .buildFileNameFilterFromWildChar(string));
        }
        RegexFileNameFilter filter = FileNamePatternUtil.getInstance()
                .buildFileNameFilterFromFilters(filterList);

        try {
            while (!getAllFiles) {

                try {
                    if (this.logger.isLoggable(Level.FINEST)) {
                        logger.finest("start getting files from the FTP server");
                    }

                    if (client == null || !client.isConnected()) {
                        client = connectToFTPDirectory();
                    }

                    for (FTPFile file : client.listFiles()) {
                        if (file.isFile() && filter.accept(null, file.getName())) {
                            datahandlers.put(file.getName(), getAsDataHandler(file, client));
                        }
                    }
                    getAllFiles = true;
                } catch (IOException e) {
                    numberOfTries++;
                    if (numberOfTries == maxTries) {
                        throw e;
                    } else {
                        this.logger.warning("An attempt to get a file failed : " + e.getMessage());
                    }
                    Thread.sleep(info.getDelay());
                }
            }
        } finally {
            if (client != null && client.isConnected()) {
                logger.finest("disconnect FTP connection to " + info.getServer());
                client.logout();
                client.disconnect();
            }
        }

        return datahandlers;
    }

    /**
     * List files from FTP server
     * 
     * @return
     * @throws IOException
     * @throws InterruptedException
     */
    public List<String> list() throws IOException, InterruptedException {

        FTPClient client = null;
        List<String> fileList = new ArrayList<String>();
        boolean listAllFiles = false;
        int numberOfTries = 0;
        int maxTries = info.getAttempt();

        try {
            while (!listAllFiles) {

                try {

                    logger.finest("start listing a directory from the FTP server");

                    if (client == null || !client.isConnected()) {
                        client = connectToFTPDirectory();
                    }

                    for (FTPFile file : client.listFiles()) {
                        if (file.isFile()) {
                            fileList.add(file.getName());
                        }
                    }
                    listAllFiles = true;
                } catch (IOException e) {
                    numberOfTries++;
                    if (numberOfTries == maxTries) {
                        throw e;
                    } else {
                        this.logger.warning("An attempt to list a directory failed : "
                                + e.getMessage());
                    }
                    Thread.sleep(info.getDelay());
                }
            }
        } finally {
            if (client != null && client.isConnected()) {
                logger.finest("disconnect FTP connection to " + info.getServer());
                client.logout();
                client.disconnect();
            }
        }

        return fileList;
    }

    /**
     * The ftp-client MUST be connected. This method DOES NOT close the
     * ftp-client
     * 
     * @param remoteFile
     * @param client
     *            a ftpclient, MUST be CONNECTED
     * @return the retrieved file, as a datahandler
     * @throws IOException
     *             the file is not found or FTP access problem
     */
    private DataHandler getAsDataHandler(FTPFile remoteFile, FTPClient client) throws IOException {
        DataHandler result;
        logger.finest("start receiving file " + remoteFile.getName() + " from FTP server");

        ByteArrayOutputStream outputStream = new ByteArrayOutputStream((int) remoteFile.getSize());
        client.retrieveFile(remoteFile.getName(), outputStream);
        outputStream.flush();
        logger.finest(remoteFile.getName() + " file received from FTP server");
        DataSource datasource = new ByteArrayDataSource(outputStream.toByteArray(), FileTypeMap
                .getDefaultFileTypeMap().getContentType(remoteFile.getName()));
        result = new DataHandler(datasource);

        logger.finest(remoteFile.getName() + " set as DataHandler");
        return result;
    }

    /**
     * The ftp-client MUST be connected. This method DOES NOT close the
     * ftp-client
     * 
     * @param remoteFile
     * @param client
     *            a ftpclient, MUST be CONNECTED
     * @return the retrieved file, as a Source
     * @throws IOException
     *             the file is not found or FTP access problem
     */
    private Source getAsSource(FTPFile remoteFile, FTPClient client) throws IOException {
        Source result;
        if (this.logger.isLoggable(Level.FINEST)) {
            logger.finest("start receiving file " + remoteFile.getName() + " from FTP server");
        }

        ByteArrayOutputStream outputStream = new ByteArrayOutputStream((int) remoteFile.getSize());
        client.retrieveFile(remoteFile.getName(), outputStream);
        outputStream.flush();
        if (this.logger.isLoggable(Level.FINEST)) {
            logger.finest(remoteFile.getName() + " file received from FTP server");
        }
        result = new StreamSource(new ByteArrayInputStream(outputStream.toByteArray()));
        if (this.logger.isLoggable(Level.FINEST)) {
            logger.finest(remoteFile.getName() + " set as Source");
        }
        return result;
    }

    /**
     * Create a FTPClient using the current FTPConnectionInfo, in the defined
     * Directory. Set active/passive mode and auto/ascii/bin transfer type
     * 
     * @return
     * @throws IOException
     */
    private FTPClient connectToFTPDirectory() throws IOException {
        logger.finest("create FTP connection to " + info.getServer());
        FTPClient client = FTPClientPool.createNewFTPClient(info.getServer(), info.getPort(),
                info.getUser(), info.getPassword());
        boolean cd = true;
        if (info.getDirectory().length() > 0) {
            cd = client.changeWorkingDirectory(info.getDirectory());
        }
        if (cd == false) {
            throw new IOException("Failed to change directory. Target directory : [" + info.getDirectory()+"]");
        }
        // connection mode
        if (info.isPassiveMode()) {
            client.enterLocalPassiveMode();
            logger.finest("enter Passive mode");
        }
        // transfer mode
        if (info.isAsciiTransferType()) {
            client.setFileType(FTP.ASCII_FILE_TYPE);
            logger.finest("ascii transfer mode enable");
        } else if (info.isBinaryTransferType()) {
            client.setFileType(FTP.BINARY_FILE_TYPE);
            logger.finest("binary transfer mode enable");
        }

        logger.finest(info.toString());
        logger.finest("FTP server working directory is : " + client.printWorkingDirectory());
        return client;

    }

    /***
     * Search if the file exist on the remote directory
     * 
     * @param srcFileName
     *            Name of the file
     * @param client
     * @return true if the file exist on the remote directory (case-sensitive)
     * @throws IOException
     */
    private boolean fileExistOnRemoteDirectory(String srcFileName, FTPClient client)
            throws IOException {
        boolean fileExist = false;
        String[] remoteFileNames = client.listNames();
        for (String remoteFileName : remoteFileNames) {
            if (remoteFileName.equals(srcFileName)) {
                fileExist = true;
                break;
            }
        }
        return fileExist;
    }
}
